@@ -1,19 +1,15 @@
-use std::usize;
-
-use ruff_formatter::{format_args, write, FormatRuleWithOptions};
-use ruff_python_ast::node::{AnyNodeRef, AstNode};
-use ruff_python_ast::{Parameters, Ranged};
-use ruff_python_trivia::{SimpleToken, SimpleTokenKind, SimpleTokenizer};
-use ruff_text_size::{TextRange, TextSize};
+use ruff_formatter::{FormatRuleWithOptions, format_args, write};
+use ruff_python_ast::{AnyNodeRef, Parameters};
+use ruff_python_trivia::{CommentLinePosition, SimpleToken, SimpleTokenKind, SimpleTokenizer};
+use ruff_text_size::{Ranged, TextRange, TextSize};
 
 use crate::comments::{
-    dangling_open_parenthesis_comments, leading_comments, leading_node_comments, trailing_comments,
-    CommentLinePosition, SourceComment,
+    SourceComment, dangling_comments, dangling_open_parenthesis_comments, leading_comments,
+    leading_node_comments, trailing_comments,
 };
 use crate::context::{NodeLevel, WithNodeLevel};
 use crate::expression::parentheses::empty_parenthesized;
 use crate::prelude::*;
-use crate::FormatNodeRule;
 
 #[derive(Debug, Copy, Clone, Eq, PartialEq, Default)]
 pub enum ParametersParentheses {
@@ -53,6 +49,7 @@ impl FormatNodeRule<Parameters> for FormatParameters {
     fn fmt_fields(&self, item: &Parameters, f: &mut PyFormatter) -> FormatResult<()> {
         let Parameters {
             range: _,
+            node_index: _,
             posonlyargs,
             args,
             vararg,
@@ -63,7 +60,7 @@ impl FormatNodeRule<Parameters> for FormatParameters {
         let (slash, star) = find_parameter_separators(f.context().source(), item);
 
         let comments = f.context().comments().clone();
-        let dangling = comments.dangling_comments(item);
+        let dangling = comments.dangling(item);
 
         // First dangling comment: trailing the opening parenthesis, e.g.:
         // ```python
@@ -103,7 +100,15 @@ impl FormatNodeRule<Parameters> for FormatParameters {
             dangling.split_at(parenthesis_comments_end);
 
         let format_inner = format_with(|f: &mut PyFormatter| {
-            let separator = format_with(|f| write!(f, [text(","), soft_line_break_or_space()]));
+            let separator = format_with(|f: &mut PyFormatter| {
+                token(",").fmt(f)?;
+
+                if f.context().node_level().is_parenthesized() {
+                    soft_line_break_or_space().fmt(f)
+                } else {
+                    space().fmt(f)
+                }
+            });
             let mut joiner = f.join_with(separator);
             let mut last_node: Option<AnyNodeRef> = None;
 
@@ -128,7 +133,7 @@ impl FormatNodeRule<Parameters> for FormatParameters {
                     let assignment = assign_argument_separator_comment_placement(
                         slash.as_ref(),
                         star.as_ref(),
-                        comment.slice().range(),
+                        comment.range(),
                         comment.line_position(),
                     )
                     .expect("Unexpected dangling comment type in function parameters");
@@ -157,10 +162,10 @@ impl FormatNodeRule<Parameters> for FormatParameters {
             if let Some(vararg) = vararg {
                 joiner.entry(&format_args![
                     leading_node_comments(vararg.as_ref()),
-                    text("*"),
+                    token("*"),
                     vararg.format()
                 ]);
-                last_node = Some(vararg.as_any_node_ref());
+                last_node = Some(vararg.as_ref().into());
             } else if !kwonlyargs.is_empty() {
                 // Given very strange comment placement, comments here may not actually have been
                 // marked as `StarLeading`/`StarTrailing`, but that's fine since we still produce
@@ -193,10 +198,10 @@ impl FormatNodeRule<Parameters> for FormatParameters {
             if let Some(kwarg) = kwarg {
                 joiner.entry(&format_args![
                     leading_node_comments(kwarg.as_ref()),
-                    text("**"),
+                    token("**"),
                     kwarg.format()
                 ]);
-                last_node = Some(kwarg.as_any_node_ref());
+                last_node = Some(kwarg.as_ref().into());
             }
 
             joiner.finish()?;
@@ -217,10 +222,10 @@ impl FormatNodeRule<Parameters> for FormatParameters {
                 // For lambdas (no parentheses), preserve the trailing comma. It doesn't
                 // behave like a magic trailing comma, it's just preserved
                 if has_trailing_comma(item, last_node, f.context().source()) {
-                    write!(f, [text(",")])?;
+                    write!(f, [token(",")])?;
                 }
             } else {
-                write!(f, [if_group_breaks(&text(","))])?;
+                write!(f, [if_group_breaks(&token(","))])?;
 
                 if f.options().magic_trailing_comma().is_respect()
                     && has_trailing_comma(item, last_node, f.context().source())
@@ -233,39 +238,43 @@ impl FormatNodeRule<Parameters> for FormatParameters {
             Ok(())
         });
 
-        let mut f = WithNodeLevel::new(NodeLevel::ParenthesizedExpression, f);
-
-        let num_parameters = posonlyargs.len()
-            + args.len()
-            + usize::from(vararg.is_some())
-            + kwonlyargs.len()
-            + usize::from(kwarg.is_some());
+        let num_parameters = item.len();
 
         if self.parentheses == ParametersParentheses::Never {
-            write!(f, [group(&format_inner)])
+            write!(f, [group(&format_inner), dangling_comments(dangling)])
         } else if num_parameters == 0 {
+            let mut f = WithNodeLevel::new(NodeLevel::ParenthesizedExpression, f);
             // No parameters, format any dangling comments between `()`
             write!(f, [empty_parenthesized("(", dangling, ")")])
+        } else if num_parameters == 1 && posonlyargs.is_empty() && kwonlyargs.is_empty() {
+            // If we have a single argument, avoid the inner group, to ensure that we insert a
+            // trailing comma if the outer group breaks.
+            let mut f = WithNodeLevel::new(NodeLevel::ParenthesizedExpression, f);
+            write!(
+                f,
+                [
+                    token("("),
+                    dangling_open_parenthesis_comments(parenthesis_dangling),
+                    soft_block_indent(&format_inner),
+                    token(")")
+                ]
+            )
         } else {
             // Intentionally avoid `parenthesized`, which groups the entire formatted contents.
             // We want parameters to be grouped alongside return types, one level up, so we
             // format them "inline" here.
+            let mut f = WithNodeLevel::new(NodeLevel::ParenthesizedExpression, f);
             write!(
                 f,
                 [
-                    text("("),
+                    token("("),
                     dangling_open_parenthesis_comments(parenthesis_dangling),
                     soft_block_indent(&group(&format_inner)),
-                    text(")")
+                    token(")")
                 ]
             )
         }
     }
-
-    fn fmt_dangling_comments(&self, _node: &Parameters, _f: &mut PyFormatter) -> FormatResult<()> {
-        // Handled in `fmt_fields`
-        Ok(())
-    }
 }
 
 struct CommentsAroundText<'a> {
@@ -276,7 +285,7 @@ struct CommentsAroundText<'a> {
 impl Format<PyFormatContext<'_>> for CommentsAroundText<'_> {
     fn fmt(&self, f: &mut PyFormatter) -> FormatResult<()> {
         if self.comments.is_empty() {
-            text(self.text).fmt(f)
+            token(self.text).fmt(f)
         } else {
             // There might be own line comments in trailing, but those are weird and we can kinda
             // ignore them
@@ -298,7 +307,7 @@ impl Format<PyFormatContext<'_>> for CommentsAroundText<'_> {
                 f,
                 [
                     leading_comments(leading),
-                    text(self.text),
+                    token(self.text),
                     trailing_comments(trailing)
                 ]
             )
@@ -412,7 +421,7 @@ pub(crate) fn find_parameter_separators(
             debug_assert!(star.kind() == SimpleTokenKind::Star, "{star:?}");
 
             Some(ParameterSeparator {
-                preceding_end: parameters.range.start(),
+                preceding_end: parameters.start(),
                 separator: star.range,
                 following_start: first_keyword_argument.start(),
             })
@@ -426,6 +435,7 @@ pub(crate) fn find_parameter_separators(
     // * `f(a, /, b)`
     // * `f(a, /, *b)`
     // * `f(a, /, *, b)`
+    // * `f(a, /, *, **b)`
     // * `f(a, /)`
     let slash_following_start = parameters
         .args
@@ -433,6 +443,7 @@ pub(crate) fn find_parameter_separators(
         .map(Ranged::start)
         .or(parameters.vararg.as_ref().map(|first| first.start()))
         .or(star.as_ref().map(|star| star.separator.start()))
+        .or(parameters.kwarg.as_deref().map(Ranged::start))
         .unwrap_or(parameters.end());
     let slash = slash.map(|(preceding_end, slash)| ParameterSeparator {
         preceding_end,
@@ -660,10 +671,28 @@ fn has_trailing_comma(
     // The slash lacks its own node
     if ends_with_pos_only_argument_separator {
         let comma = tokens.next();
-        assert!(matches!(comma, Some(SimpleToken { kind: SimpleTokenKind::Comma, .. })), "The last positional only argument must be separated by a `,` from the positional only parameters separator `/` but found '{comma:?}'.");
+        assert!(
+            matches!(
+                comma,
+                Some(SimpleToken {
+                    kind: SimpleTokenKind::Comma,
+                    ..
+                })
+            ),
+            "The last positional only argument must be separated by a `,` from the positional only parameters separator `/` but found '{comma:?}'."
+        );
 
         let slash = tokens.next();
-        assert!(matches!(slash, Some(SimpleToken { kind: SimpleTokenKind::Slash, .. })), "The positional argument separator must be present for a function that has positional only parameters but found '{slash:?}'.");
+        assert!(
+            matches!(
+                slash,
+                Some(SimpleToken {
+                    kind: SimpleTokenKind::Slash,
+                    ..
+                })
+            ),
+            "The positional argument separator must be present for a function that has positional only parameters but found '{slash:?}'."
+        );
     }
 
     tokens
