@@ -1,7 +1,11 @@
 //! Data model, state management, and configuration resolution.
 
+use std::collections::{BTreeMap, HashSet, VecDeque};
+use std::ops::{Deref, DerefMut};
+use std::panic::RefUnwindSafe;
+use std::sync::Arc;
+
 use anyhow::{Context, anyhow};
-use index::DocumentError;
 use lsp_server::{Message, RequestId};
 use lsp_types::notification::{DidChangeWatchedFiles, Exit, Notification};
 use lsp_types::request::{
@@ -13,20 +17,17 @@ use lsp_types::{
     DidChangeWatchedFilesRegistrationOptions, FileSystemWatcher, Registration, RegistrationParams,
     TextDocumentContentChangeEvent, Unregistration, UnregistrationParams, Url,
 };
-use options::GlobalOptions;
 use ruff_db::Db;
 use ruff_db::files::{File, system_path_to_file};
 use ruff_db::system::{System, SystemPath, SystemPathBuf};
-use std::borrow::Cow;
-use std::collections::{BTreeMap, HashSet, VecDeque};
-use std::ops::{Deref, DerefMut};
-use std::panic::RefUnwindSafe;
-use std::sync::Arc;
 use ty_combine::Combine;
 use ty_project::metadata::Options;
-use ty_project::watch::ChangeEvent;
+use ty_project::watch::{ChangeEvent, CreatedKind};
 use ty_project::{ChangeResult, CheckMode, Db as _, ProjectDatabase, ProjectMetadata};
 
+use index::DocumentError;
+use options::GlobalOptions;
+
 pub(crate) use self::options::InitializationOptions;
 pub use self::options::{ClientOptions, DiagnosticMode};
 pub(crate) use self::settings::{GlobalSettings, WorkspaceSettings};
@@ -36,6 +37,7 @@ use crate::capabilities::{
 use crate::document::{DocumentKey, DocumentVersion, NotebookDocument};
 use crate::server::{Action, publish_settings_diagnostics};
 use crate::session::client::Client;
+use crate::session::index::Document;
 use crate::session::request_queue::RequestQueue;
 use crate::system::{AnySystemPath, LSPSystem};
 use crate::{PositionEncoding, TextDocument};
@@ -816,25 +818,16 @@ impl Session {
         let index = self.index();
         let document_handle = index.document_handle(url)?;
 
-        let notebook = if let Some(notebook_path) = &document_handle.notebook_path {
-            index
-                .notebook_arc(&DocumentKey::from(notebook_path.clone()))
-                .ok()
-        } else {
-            None
-        };
-
         Ok(DocumentSnapshot {
             resolved_client_capabilities: self.resolved_client_capabilities,
             global_settings: self.global_settings.clone(),
             workspace_settings: document_handle
-                .to_file_path()
+                .notebook_or_file_path()
                 .as_system()
                 .and_then(|path| self.workspaces.settings_for_path(path))
                 .unwrap_or_else(|| Arc::new(WorkspaceSettings::default())),
             position_encoding: self.position_encoding,
             document: document_handle,
-            notebook,
         })
     }
 
@@ -860,13 +853,7 @@ impl Session {
     pub(super) fn text_document_handles(&self) -> impl Iterator<Item = DocumentHandle> + '_ {
         self.index()
             .text_documents()
-            .map(|(key, document)| DocumentHandle {
-                key: key.clone(),
-                url: document.url().clone(),
-                version: document.version(),
-                // TODO: Set notebook path if text document is part of a notebook
-                notebook_path: None,
-            })
+            .map(|(_, document)| DocumentHandle::from_text_document(document))
     }
 
     /// Returns a handle to the document specified by its URL.
@@ -887,7 +874,7 @@ impl Session {
     /// Returns a handle to the opened document.
     pub(crate) fn open_notebook_document(&mut self, document: NotebookDocument) -> DocumentHandle {
         let handle = self.index_mut().open_notebook_document(document);
-        self.bump_revision();
+        self.open_document_in_db(&handle);
         handle
     }
 
@@ -897,9 +884,49 @@ impl Session {
     /// Returns a handle to the opened document.
     pub(crate) fn open_text_document(&mut self, document: TextDocument) -> DocumentHandle {
         let handle = self.index_mut().open_text_document(document);
+        self.open_document_in_db(&handle);
+        handle
+    }
+
+    fn open_document_in_db(&mut self, document: &DocumentHandle) {
+        let path = document.notebook_or_file_path();
+
+        // This is a "maybe" because the `File` might've not been interned yet i.e., the
+        // `try_system` call will return `None` which doesn't mean that the file is new, it's just
+        // that the server didn't need the file yet.
+        let is_maybe_new_system_file = path.as_system().is_some_and(|system_path| {
+            let db = self.project_db(path);
+            db.files()
+                .try_system(db, system_path)
+                .is_none_or(|file| !file.exists(db))
+        });
+
+        match path {
+            AnySystemPath::System(system_path) => {
+                let event = if is_maybe_new_system_file {
+                    ChangeEvent::Created {
+                        path: system_path.clone(),
+                        kind: CreatedKind::File,
+                    }
+                } else {
+                    ChangeEvent::Opened(system_path.clone())
+                };
+                self.apply_changes(path, vec![event]);
+
+                let db = self.project_db_mut(path);
+                match system_path_to_file(db, system_path) {
+                    Ok(file) => db.project().open_file(db, file),
+                    Err(err) => tracing::warn!("Failed to open file {system_path}: {err}"),
+                }
+            }
+            AnySystemPath::SystemVirtual(virtual_path) => {
+                let db = self.project_db_mut(path);
+                let virtual_file = db.files().virtual_file(db, virtual_path);
+                db.project().open_file(db, virtual_file.file());
+            }
+        }
 
         self.bump_revision();
-        handle
     }
 
     /// Returns a reference to the index.
@@ -999,7 +1026,6 @@ pub(crate) struct DocumentSnapshot {
     workspace_settings: Arc<WorkspaceSettings>,
     position_encoding: PositionEncoding,
     document: DocumentHandle,
-    notebook: Option<Arc<NotebookDocument>>,
 }
 
 impl DocumentSnapshot {
@@ -1028,17 +1054,12 @@ impl DocumentSnapshot {
         &self.document
     }
 
-    /// Returns the URL of the document.
     pub(crate) fn url(&self) -> &lsp_types::Url {
         self.document.url()
     }
 
-    pub(crate) fn notebook(&self) -> Option<&NotebookDocument> {
-        self.notebook.as_deref()
-    }
-
-    pub(crate) fn to_file(&self, db: &dyn Db) -> Option<File> {
-        let file = self.document.to_file(db);
+    pub(crate) fn to_notebook_or_file(&self, db: &dyn Db) -> Option<File> {
+        let file = self.document.notebook_or_file(db);
         if file.is_none() {
             tracing::debug!(
                 "Failed to resolve file: file not found for `{}`",
@@ -1048,8 +1069,8 @@ impl DocumentSnapshot {
         file
     }
 
-    pub(crate) fn to_file_path(&self) -> Cow<'_, AnySystemPath> {
-        self.document.to_file_path()
+    pub(crate) fn notebook_or_file_path(&self) -> &AnySystemPath {
+        self.document.notebook_or_file_path()
     }
 }
 
@@ -1330,34 +1351,99 @@ impl SuspendedWorkspaceDiagnosticRequest {
 ///
 /// It also exposes methods to get the file-path of the corresponding ty-file.
 #[derive(Clone, Debug)]
-pub(crate) struct DocumentHandle {
-    /// The key that uniquely identifies this document in the index.
-    key: DocumentKey,
-    url: lsp_types::Url,
-    /// The path to the enclosing notebook file if this document is a notebook or a notebook cell.
-    notebook_path: Option<AnySystemPath>,
-    version: DocumentVersion,
+pub(crate) enum DocumentHandle {
+    Text {
+        url: lsp_types::Url,
+        path: AnySystemPath,
+        version: DocumentVersion,
+    },
+    Notebook {
+        url: lsp_types::Url,
+        path: AnySystemPath,
+        version: DocumentVersion,
+    },
+    Cell {
+        url: lsp_types::Url,
+        version: DocumentVersion,
+        notebook_path: AnySystemPath,
+    },
 }
 
 impl DocumentHandle {
+    fn from_text_document(document: &TextDocument) -> Self {
+        match document.notebook() {
+            None => Self::Text {
+                version: document.version(),
+                url: document.url().clone(),
+                path: DocumentKey::from_url(document.url()).into_file_path(),
+            },
+            Some(notebook) => Self::Cell {
+                notebook_path: notebook.clone(),
+                version: document.version(),
+                url: document.url().clone(),
+            },
+        }
+    }
+
+    fn from_notebook_document(document: &NotebookDocument) -> Self {
+        Self::Notebook {
+            path: DocumentKey::from_url(document.url()).into_file_path(),
+            url: document.url().clone(),
+            version: document.version(),
+        }
+    }
+
+    fn from_document(document: &Document) -> Self {
+        match document {
+            Document::Text(text) => Self::from_text_document(text),
+            Document::Notebook(notebook) => Self::from_notebook_document(notebook),
+        }
+    }
+
+    fn key(&self) -> DocumentKey {
+        DocumentKey::from_url(self.url())
+    }
+
     pub(crate) const fn version(&self) -> DocumentVersion {
-        self.version
+        match self {
+            Self::Text { version, .. }
+            | Self::Notebook { version, .. }
+            | Self::Cell { version, .. } => *version,
+        }
     }
 
     /// The URL as used by the client to reference this document.
     pub(crate) fn url(&self) -> &lsp_types::Url {
-        &self.url
+        match self {
+            Self::Text { url, .. } | Self::Notebook { url, .. } | Self::Cell { url, .. } => url,
+        }
     }
 
     /// The path to the enclosing file for this document.
     ///
     /// This is the path corresponding to the URL, except for notebook cells where the
     /// path corresponds to the notebook file.
-    pub(crate) fn to_file_path(&self) -> Cow<'_, AnySystemPath> {
-        if let Some(path) = self.notebook_path.as_ref() {
-            Cow::Borrowed(path)
-        } else {
-            Cow::Owned(self.key.to_file_path())
+    pub(crate) fn notebook_or_file_path(&self) -> &AnySystemPath {
+        match self {
+            Self::Text { path, .. } | Self::Notebook { path, .. } => path,
+            Self::Cell { notebook_path, .. } => notebook_path,
+        }
+    }
+
+    #[expect(unused)]
+    pub(crate) fn file_path(&self) -> Option<&AnySystemPath> {
+        match self {
+            Self::Text { path, .. } | Self::Notebook { path, .. } => Some(path),
+            Self::Cell { .. } => None,
+        }
+    }
+
+    #[expect(unused)]
+    pub(crate) fn notebook_path(&self) -> Option<&AnySystemPath> {
+        match self {
+            DocumentHandle::Notebook { path, .. } => Some(path),
+            DocumentHandle::Cell { notebook_path, .. } => Some(notebook_path),
+            DocumentHandle::Text { .. } => None,
         }
     }
 
@@ -1366,8 +1452,8 @@ impl DocumentHandle {
     /// It returns [`None`] for the following cases:
     /// - For virtual file, if it's not yet opened
     /// - For regular file, if it does not exists or is a directory
-    pub(crate) fn to_file(&self, db: &dyn Db) -> Option<File> {
-        match &*self.to_file_path() {
+    pub(crate) fn notebook_or_file(&self, db: &dyn Db) -> Option<File> {
+        match &self.notebook_or_file_path() {
             AnySystemPath::System(path) => system_path_to_file(db, path).ok(),
             AnySystemPath::SystemVirtual(virtual_path) => db
                 .files()
@@ -1376,6 +1462,14 @@ impl DocumentHandle {
         }
     }
 
+    pub(crate) fn is_cell(&self) -> bool {
+        matches!(self, Self::Cell { .. })
+    }
+
+    pub(crate) fn is_cell_or_notebook(&self) -> bool {
+        matches!(self, Self::Cell { .. } | Self::Notebook { .. })
+    }
+
     pub(crate) fn update_text_document(
         &self,
         session: &mut Session,
@@ -1383,29 +1477,121 @@ impl DocumentHandle {
         new_version: DocumentVersion,
     ) -> crate::Result<()> {
         let position_encoding = session.position_encoding();
-        let mut index = session.index_mut();
+        {
+            let mut index = session.index_mut();
 
-        let document_mut = index.document_mut(&self.key)?;
+            let document_mut = index.document_mut(&self.key())?;
 
-        let Some(document) = document_mut.as_text_mut() else {
-            anyhow::bail!("Text document path does not point to a text document");
-        };
+            let Some(document) = document_mut.as_text_mut() else {
+                anyhow::bail!("Text document path does not point to a text document");
+            };
 
-        if content_changes.is_empty() {
-            document.update_version(new_version);
-            return Ok(());
+            if content_changes.is_empty() {
+                document.update_version(new_version);
+            } else {
+                document.apply_changes(content_changes, new_version, position_encoding);
+            }
         }
 
-        document.apply_changes(content_changes, new_version, position_encoding);
+        self.update_in_db(session);
 
         Ok(())
     }
 
+    pub(crate) fn update_notebook_document(
+        &self,
+        session: &mut Session,
+        cells: Option<lsp_types::NotebookDocumentCellChange>,
+        metadata: Option<lsp_types::LSPObject>,
+        new_version: DocumentVersion,
+    ) -> crate::Result<()> {
+        let position_encoding = session.position_encoding();
+        {
+            let mut index = session.index_mut();
+
+            index.update_notebook_document(
+                &self.key(),
+                cells,
+                metadata,
+                new_version,
+                position_encoding,
+            )?;
+        }
+
+        self.update_in_db(session);
+        Ok(())
+    }
+
+    fn update_in_db(&self, session: &mut Session) {
+        let path = self.notebook_or_file_path();
+        let changes = match path {
+            AnySystemPath::System(system_path) => {
+                vec![ChangeEvent::file_content_changed(system_path.clone())]
+            }
+            AnySystemPath::SystemVirtual(virtual_path) => {
+                vec![ChangeEvent::ChangedVirtual(virtual_path.clone())]
+            }
+        };
+
+        session.apply_changes(path, changes);
+    }
+
     /// De-registers a document, specified by its key.
     /// Calling this multiple times for the same document is a logic error.
-    pub(crate) fn close(self, session: &mut Session) -> crate::Result<()> {
-        session.index_mut().close_document(&self.key)?;
+    ///
+    /// Returns `true` if the client needs to clear the diagnostics for this document.
+    pub(crate) fn close(&self, session: &mut Session) -> crate::Result<bool> {
+        let is_cell = self.is_cell();
+        let path = self.notebook_or_file_path();
+        session.index_mut().close_document(&self.key())?;
+
+        // Close the text or notebook file in the database but skip this
+        // step for cells because closing a cell doesn't close its notebook.
+        let requires_clear_diagnostics = if is_cell {
+            true
+        } else {
+            let db = session.project_db_mut(path);
+
+            match path {
+                AnySystemPath::System(system_path) => {
+                    if let Some(file) = db.files().try_system(db, system_path) {
+                        db.project().close_file(db, file);
+                    } else {
+                        // This can only fail when the path is a directory or it doesn't exists but the
+                        // file should exists for this handler in this branch. This is because every
+                        // close call is preceded by an open call, which ensures that the file is
+                        // interned in the lookup table (`Files`).
+                        tracing::warn!("Salsa file does not exists for {}", system_path);
+                    }
+
+                    // For non-virtual files, we clear diagnostics if:
+                    //
+                    // 1. The file does not belong to any workspace e.g., opening a random file from
+                    //    outside the workspace because closing it acts like the file doesn't exists
+                    // 2. The diagnostic mode is set to open-files only
+                    session.workspaces().for_path(system_path).is_none()
+                        || session
+                            .global_settings()
+                            .diagnostic_mode()
+                            .is_open_files_only()
+                }
+                AnySystemPath::SystemVirtual(virtual_path) => {
+                    if let Some(virtual_file) = db.files().try_virtual_file(virtual_path) {
+                        db.project().close_file(db, virtual_file.file());
+                        virtual_file.close(db);
+                    } else {
+                        tracing::warn!("Salsa virtual file does not exists for {}", virtual_path);
+                    }
+
+                    // Always clear diagnostics for virtual files, as they don't really exist on disk
+                    // which means closing them is like deleting the file.
+                    true
+                }
+            }
+        };
+
         session.bump_revision();
-        Ok(())
+
+        Ok(requires_clear_diagnostics)
     }
 }
