@@ -30,11 +30,12 @@
 mod commands;
 mod initialize;
 mod inlay_hints;
+mod notebook;
 mod publish_diagnostics;
 mod pull_diagnostics;
 
 use std::collections::hash_map::Entry;
-use std::collections::{HashMap, VecDeque};
+use std::collections::{BTreeMap, HashMap, VecDeque};
 use std::num::NonZeroUsize;
 use std::sync::{Arc, OnceLock};
 use std::thread::JoinHandle;
@@ -166,6 +167,8 @@ impl TestServer {
     ) -> Result<Self> {
         setup_tracing();
 
+        tracing::debug!("Starting test client with capabilities {:#?}", capabilities);
+
         let (server_connection, client_connection) = Connection::memory();
 
         // Create OS system with the test directory as cwd
@@ -345,6 +348,7 @@ impl TestServer {
         }
 
         let id = self.next_request_id();
+        tracing::debug!("Client sends request `{}` with ID {}", R::METHOD, id);
         let request = lsp_server::Request::new(id.clone(), R::METHOD.to_string(), params);
         self.send(Message::Request(request));
         id
@@ -356,6 +360,7 @@ impl TestServer {
         N: Notification,
     {
         let notification = lsp_server::Notification::new(N::METHOD.to_string(), params);
+        tracing::debug!("Client sends notification `{}`", N::METHOD);
         self.send(Message::Notification(notification));
     }
 
@@ -433,6 +438,33 @@ impl TestServer {
         ))
     }
 
+    /// Collects `N` publish diagnostic notifications into a map, indexed by the document url.
+    ///
+    /// ## Panics
+    /// If there are multiple publish diagnostics notifications for the same document.
+    pub(crate) fn collect_publish_diagnostic_notifications(
+        &mut self,
+        count: usize,
+    ) -> Result<BTreeMap<lsp_types::Url, Vec<lsp_types::Diagnostic>>> {
+        let mut results = BTreeMap::default();
+
+        for _ in 0..count {
+            let notification =
+                self.await_notification::<lsp_types::notification::PublishDiagnostics>()?;
+
+            if let Some(existing) =
+                results.insert(notification.uri.clone(), notification.diagnostics)
+            {
+                panic!(
+                    "Received multiple publish diagnostic notifications for {url}: ({existing:#?})",
+                    url = &notification.uri
+                );
+            }
+        }
+
+        Ok(results)
+    }
+
     /// Wait for a request of the specified type from the server and return the request ID and
     /// parameters.
     ///
@@ -512,7 +544,7 @@ impl TestServer {
     fn handle_message(&mut self, message: Message) -> Result<(), TestServerError> {
         match message {
             Message::Request(request) => {
-                tracing::debug!("Received server request {}", &request.method);
+                tracing::debug!("Received server request `{}`", &request.method);
                 self.requests.push_back(request);
             }
             Message::Response(response) => {
@@ -530,7 +562,7 @@ impl TestServer {
                 }
             }
             Message::Notification(notification) => {
-                tracing::debug!("Received notification {}", &notification.method);
+                tracing::debug!("Received notification `{}`", &notification.method);
                 self.notifications.push_back(notification);
             }
         }
@@ -774,6 +806,7 @@ impl Drop for TestServer {
             match self.await_response::<Shutdown>(&shutdown_id) {
                 Ok(()) => {
                     self.send_notification::<Exit>(());
+
                     None
                 }
                 Err(err) => Some(format!("Failed to get shutdown response: {err:?}")),
@@ -782,9 +815,32 @@ impl Drop for TestServer {
             None
         };
 
-        if let Some(_client_connection) = self.client_connection.take() {
-            // Drop the client connection before joining the server thread to avoid any hangs
-            // in case the server didn't respond to the shutdown request.
+        // Drop the client connection before joining the server thread to avoid any hangs
+        // in case the server didn't respond to the shutdown request.
+        if let Some(client_connection) = self.client_connection.take() {
+            if !std::thread::panicking() {
+                // Wait for the client sender to drop (confirmation that it processed the exit notification).
+
+                match client_connection
+                    .receiver
+                    .recv_timeout(Duration::from_secs(20))
+                {
+                    Err(RecvTimeoutError::Disconnected) => {
+                        // Good, the server terminated
+                    }
+                    Err(RecvTimeoutError::Timeout) => {
+                        tracing::warn!(
+                            "The server didn't exit within 20ms after receiving the EXIT notification"
+                        );
+                    }
+                    Ok(message) => {
+                        // Ignore any errors: A duplicate pending message
+                        // won't matter that much because `assert_no_pending_messages` will
+                        // panic anyway.
+                        let _ = self.handle_message(message);
+                    }
+                }
+            }
         }
 
         if std::thread::panicking() {
