@@ -48,13 +48,16 @@ impl EnvVars {
     /// (e.g. colons on Unix or semicolons on Windows).
     pub const PYTHONPATH: &'static str = "PYTHONPATH";
 
-    /// Used to determine if an active Conda environment is the base environment or not.
+    /// Used to determine the name of the active Conda environment.
     pub const CONDA_DEFAULT_ENV: &'static str = "CONDA_DEFAULT_ENV";
 
-    /// Used to detect an activated Conda environment location.
+    /// Used to detect the path of an active Conda environment.
     /// If both `VIRTUAL_ENV` and `CONDA_PREFIX` are present, `VIRTUAL_ENV` will be preferred.
     pub const CONDA_PREFIX: &'static str = "CONDA_PREFIX";
 
+    /// Used to determine the root install path of Conda.
+    pub const CONDA_ROOT: &'static str = "_CONDA_ROOT";
+
     /// Filter which tests to run in mdtest.
     ///
     /// Only tests whose names contain this filter string will be executed.
